use std::collections::HashSet;
use std::ops::Range;

const FILE_STR: &'static str = include_str!("input.txt");

#[derive(Debug, Default, Clone, Eq, PartialEq)]
struct NumberLocation {
    start: usize,
    end: usize,
    digits: String,
}

impl NumberLocation {
    fn new(start: usize, value: char) -> Self {
        NumberLocation {
            start,
            end: start + 1,
            digits: String::from(value),
        }
    }

    fn add_digit(&mut self, digit: char) {
        self.end += 1;
        self.digits.push(digit);
    }

    fn to_number(&self) -> usize {
        self.digits.parse::<usize>().unwrap()
    }
}

#[derive(Debug, Copy, Clone, PartialOrd, PartialEq, Eq, Hash)]
enum ValueType {
    Number(usize),
    Symbol(char),
    Empty,
}

#[derive(Debug, Default, Clone)]
struct Grid {
    line_size: usize,
    numbers: Vec<NumberLocation>,
    values: Vec<ValueType>,
}

impl Grid {
    pub fn parse(input_str: &str) -> Self {
        use ValueType::*;

        let len = input_str.split('\n').next().unwrap().len();

        let mut values: Vec<ValueType> = Vec::new();
        let mut numbers: Vec<NumberLocation> = Vec::new();

        input_str
            .split('\n')
            .map(|row| row.chars())
            .enumerate()
            .for_each(|(r, char_row)| {
                let mut prev = Empty;
                char_row.enumerate().for_each(|(c, ch)| {
                    let n = len * r + c;
                    if ch.is_digit(10) {
                        if let Number(_) = prev {
                            let final_num = { numbers.len() - 1 };
                            numbers.get_mut(final_num).unwrap().add_digit(ch);
                        } else {
                            let nl = NumberLocation::new(n, ch);
                            numbers.push(nl);
                        }

                        prev = Number(numbers.len() - 1);
                    } else if ch == '.' {
                        prev = Empty;
                    } else {
                        prev = Symbol(ch);
                    }

                    values.push(prev);
                });
            });

        Grid {
            line_size: len,
            numbers,
            values,
        }
    }

    pub fn num_cols(&self) -> usize {
        self.line_size
    }

    pub fn num_rows(&self) -> usize {
        self.values.len() / self.num_cols()
    }

    // Convert x,y coordinate into linear array index
    pub fn xy_idx(&self, x: usize, y: usize) -> usize {
        (y * self.num_cols()) + x
    }

    /// Convert linear array index to x,y coordinate
    pub fn idx_xy(&self, idx: usize) -> (usize, usize) {
        (idx % self.num_cols(), idx / self.num_cols())
    }

    fn adjacent_ind_range(&self, r: Range<usize>) -> HashSet<usize> {
        let min = r.start;
        let max = r.end - 1;
        let mut ind = HashSet::new();
        if max >= self.values.len() {
            return ind;
        }

        let (min_col, row) = self.idx_xy(min);
        let (max_col, erow) = self.idx_xy(max);
        assert_eq!(row, erow, "All the digits should be in the same row");

        // Forwards/backwards
        if min_col > 0 {
            ind.insert(self.xy_idx(min_col - 1, row));
        }
        if max_col + 1 < self.num_cols() {
            ind.insert(self.xy_idx(max_col + 1, row));
        }

        // Row above/below
        for x in r {
            let (col, row) = self.idx_xy(x);

            if row + 1 < self.num_rows() {
                ind.insert(self.xy_idx(col, row + 1));
            }
            if row > 0 {
                ind.insert(self.xy_idx(col, row - 1));
            }
        }

        // Diagonals (min-1, y-1),(min-1,y+1),(max+1, y-1),(max+1, y+1)
        if min_col > 0 && row > 0 {
            ind.insert(self.xy_idx(min_col - 1, row - 1));
        }
        if min_col > 0 && row + 1 < self.num_rows() {
            ind.insert(self.xy_idx(min_col - 1, row + 1));
        }
        if max_col + 1 < self.num_cols() && row > 0 {
            ind.insert(self.xy_idx(max_col + 1, row - 1));
        }
        if max_col + 1 < self.num_cols() && row + 1 < self.num_rows() {
            ind.insert(self.xy_idx(max_col + 1, row + 1));
        }

        ind
    }

    fn find_adjacent_range(&self, r: Range<usize>) -> Vec<ValueType> {
        self.adjacent_ind_range(r)
            .into_iter()
            .map(|ind| self.values[ind])
            .collect()
    }

    fn find_adjacent(&self, i: usize) -> HashSet<ValueType> {
        HashSet::from_iter(self.find_adjacent_range(i..i + 1).iter().cloned())
    }

    fn is_part_number(&self, r: Range<usize>) -> bool {
        use ValueType::*;
        self.find_adjacent_range(r).into_iter().any(|t| match t {
            Symbol(_) => true,
            _ => false,
        })
    }

    fn get_part_numbers(&self) -> Vec<usize> {
        self.numbers
            .iter()
            .filter(|nl| {
                let range = nl.start..nl.end;
                self.is_part_number(range)
            })
            .map(|nl| nl.to_number())
            .collect()
    }

    fn get_part_number_ind(&self) -> Vec<usize> {
        self.numbers
            .iter()
            .enumerate()
            .filter(|(_, nl)| {
                let range = nl.start..nl.end;
                self.is_part_number(range)
            })
            .map(|(id, _)| id)
            .collect()
    }

    fn get_gears(&self) -> Vec<usize> {
        self.values
            .clone()
            .into_iter()
            .enumerate()
            .filter(|(_, vt)| match vt {
                ValueType::Symbol('*') => true,
                _ => false,
            })
            .map(|(id, _)| id)
            .collect()
    }

    fn get_gear_ratios(&self) -> Vec<(usize, usize)> {
        let valid_part_number_ind = self.get_part_number_ind();
        let potential_gear_symbols = self.get_gears();

        // Valid gear ratios are two valid part numbers adjacent to the gear symbol
        let raw_gear_ratios: Vec<_> = potential_gear_symbols
            .into_iter()
            .filter_map(|s| {
                let mut number_count = 0;
                let gr: Vec<_> = self
                    .find_adjacent(s)
                    .into_iter()
                    .filter(|vt| match vt {
                        ValueType::Number(id) => valid_part_number_ind.contains(id),
                        _ => false,
                    })
                    .collect();

                gr.iter().for_each(|vt| match vt {
                    ValueType::Number(_) => number_count += 1,
                    _ => {}
                });

                match number_count {
                    2 => Some(gr),
                    _ => None,
                }
            })
            .collect();

        raw_gear_ratios
            .into_iter()
            .map(|v| {
                v.into_iter()
                    .map(|vt| match vt {
                        ValueType::Number(id) => self.numbers[id].to_number(),
                        _ => panic!("Invalid gear ratio!"),
                    })
                    .collect::<Vec<usize>>()
            })
            .map(|gr| (gr[0], gr[1]))
            .collect()
    }

    fn get_part_number_sum(&self) -> usize {
        self.get_part_numbers().iter().sum()
    }

    fn get_gear_ratio_sum(&self) -> usize {
        self.get_gear_ratios()
            .into_iter()
            .map(|(gr0, gr1)| gr0 * gr1)
            .sum()
    }
}

fn part_one() {
    let grid = Grid::parse(FILE_STR);
    println!(
        "Part 1: Sum of part numbers: {}",
        grid.get_part_number_sum()
    );
}

fn part_two() {
    let grid = Grid::parse(FILE_STR);
    println!("Part 2: Sum of gear ratios: {}", grid.get_gear_ratio_sum())
}

fn main() {
    part_one();
    part_two();
}

#[cfg(test)]
mod tests {
    const EXAMPLE_FILE_STR: &'static str = include_str!("example_input.txt");
    use super::*;
    use FromIterator;

    #[test]
    fn test_get_part_number_sum_part_1() {
        let grid = Grid::parse(FILE_STR);
        assert_eq!(grid.get_part_number_sum(), 525181);
    }

    #[test]
    fn test_get_gear_ratio_sum_part_2() {
        let grid = Grid::parse(FILE_STR);
        assert_eq!(grid.get_gear_ratio_sum(), 84289137);
    }

    #[test]
    fn test_adjacent_ind_range() {
        let grid = Grid::parse(EXAMPLE_FILE_STR);
        let actual = grid.adjacent_ind_range(62..65);
        assert_eq!(12, actual.len());
        assert_eq!(
            HashSet::from_iter(actual.iter().cloned()),
            HashSet::from([51, 52, 53, 54, 55, 61, 65, 71, 72, 73, 74, 75])
        );
    }

    #[test]
    fn test_get_part_numbers() {
        let grid = Grid::parse(EXAMPLE_FILE_STR);
        let expected = [467usize, 35, 633, 617, 592, 755, 664, 598];
        assert_eq!(grid.get_part_numbers(), expected);
    }

    #[test]
    fn test_part_number_sum() {
        let grid = Grid::parse(EXAMPLE_FILE_STR);
        assert_eq!(grid.get_part_number_sum(), 4361);
    }

    #[test]
    fn test_get_gear_ratio_sum() {
        let grid = Grid::parse(EXAMPLE_FILE_STR);
        let sum = grid.get_gear_ratio_sum();
        assert_eq!(sum, 467835);
    }
}
