use std::collections::{HashMap, HashSet};
use std::iter::FromIterator;

const FILE_STR: &'static str = include_str!("input.txt");

#[derive(Debug, Default)]
struct ScratchCard {
    winning: Vec<usize>,
    chosen: Vec<usize>,
    chosen_winners: HashSet<usize>,
    cw_count: usize,
}

impl ScratchCard {
    fn new(winning: Vec<usize>, chosen: Vec<usize>) -> Self {
        ScratchCard {
            winning,
            chosen,
            ..Self::default()
        }
    }

    fn find_chosen_winners(&mut self) {
        // We've done this already
        if self.cw_count > 0 {
            return;
        }

        self.winning
            .iter()
            .filter(|w| self.chosen.contains(w))
            .for_each(|cw| _ = self.chosen_winners.insert(*cw));

        self.cw_count = self.chosen_winners.len();
    }

    fn score(&mut self) -> usize {
        self.find_chosen_winners();
        if self.chosen_winners.is_empty() {
            return 0;
        }

        let winner_count = self.chosen_winners.len() as u32;
        usize::pow(2, winner_count - 1)
    }
}

#[derive(Debug, Default)]
struct GameMap(HashMap<usize, ScratchCard>);

impl GameMap {
    fn parse(table: &str) -> Self {
        Self(HashMap::from_iter(
            table
                .split('\n')
                .filter(|s| s.len() > 0)
                .map(parse_game_table_line),
        ))
    }

    fn get_original_card_matches(&mut self) -> HashMap<usize, usize> {
        HashMap::from_iter(self.0.iter_mut().map(|(id, sc)| {
            sc.find_chosen_winners();
            (*id, sc.cw_count)
        }))
    }

    fn get_score_sum(&mut self) -> usize {
        self.0.iter_mut().map(|(_, sc)| sc.score()).sum()
    }

    fn get_scratchcard_counts(&mut self) -> HashMap<usize, usize> {
        let match_map: HashMap<usize, usize> = self.get_original_card_matches();
        let mut copy_map: HashMap<usize, usize> = HashMap::new();
        let mut count_map: HashMap<usize, usize> =
            match_map.iter().map(|(id, _)| (*id, 1usize)).collect();

        for id in 1..=match_map.len() {
            let raw_match_count = match_map.get(&id);
            if raw_match_count.is_none() {
                break;
            }
            let match_count = *raw_match_count.unwrap();
            let copy_count = match copy_map.get(&id) {
                Some(n) => *n + 1,
                None => 1,
            };

            let add_range = id + 1..=id + match_count;
            for n in add_range {
                let raw_count = count_map.get(&n);
                if let Some(count) = raw_count {
                    count_map.insert(n, *count + copy_count);
                    let new_copy_count = match copy_map.get(&n) {
                        Some(num) => *num + copy_count,
                        None => copy_count,
                    };
                    copy_map.insert(n, new_copy_count);
                } else {
                    // No value means the value of n is greater than
                    // the amount of original cards
                    break;
                }
            }
        }

        count_map
    }

    fn get_total_scratchcard_count(&mut self) -> usize {
        self.get_scratchcard_counts()
            .into_iter()
            .map(|(_, count)| count)
            .sum()
    }
}

fn parse_game_table_line(line: &str) -> (usize, ScratchCard) {
    let parts: Vec<&str> = line.split(": ").collect();
    let card_num_parts = parts[0].replace("Card", "");
    let card_num = parse_num_str(&card_num_parts)[0];
    let nums: Vec<Vec<usize>> = parts[1].split(" | ").map(parse_num_str).collect();

    (card_num, ScratchCard::new(nums[0].clone(), nums[1].clone()))
}

/// Parse numbers from a line, separated/surrounded by whitespace
fn parse_num_str(s: &str) -> Vec<usize> {
    s.split_ascii_whitespace()
        .filter(|s| s.len() > 0)
        .map(|s| s.trim())
        .map(|s| usize::from_str_radix(s, 10).unwrap())
        .collect::<Vec<usize>>()
}

fn part_one() {
    let mut map = GameMap::parse(FILE_STR);
    let sum = map.get_score_sum();
    println!("Part 1 Score Sum {}", sum);
}

fn part_two() {
    let mut map = GameMap::parse(FILE_STR);
    let card_count = map.get_total_scratchcard_count();
    println!("Part 2 Scratchcard Count: {}", card_count);
}

fn main() {
    part_one();
    part_two();
}

#[cfg(test)]
mod tests {
    use super::*;
    const EXAMPLE_FILE_STR: &'static str = include_str!("example_input.txt");

    #[test]
    fn test_part1_example() {
        let mut map = GameMap::parse(EXAMPLE_FILE_STR);
        let actual = map.get_score_sum();
        assert_eq!(13, actual);
    }

    #[test]
    fn test_part2_example() {
        let mut map = GameMap::parse(EXAMPLE_FILE_STR);
        let actual = map.get_total_scratchcard_count();
        assert_eq!(30, actual);
    }

    #[test]
    fn test_example_card_counts() {
        let mut map = GameMap::parse(EXAMPLE_FILE_STR);
        let count_map = map.get_scratchcard_counts();
        assert_eq!(Some(1), count_map.get(&1).copied());
        assert_eq!(Some(2), count_map.get(&2).copied());
        assert_eq!(Some(4), count_map.get(&3).copied());
        assert_eq!(Some(8), count_map.get(&4).copied());
        assert_eq!(Some(14), count_map.get(&5).copied());
        assert_eq!(Some(1), count_map.get(&6).copied());
    }
}
